from dagster_graphql.test.utils import execute_dagster_graphql

from .setup import define_test_context, define_test_snapshot_context

RUNTIME_TYPE_QUERY = '''
query RuntimeTypeQuery($pipelineName: String! $runtimeTypeName: String!)
{
    pipelineOrError(params:{ name: $pipelineName}) {
        __typename
        ... on Pipeline {
            runtimeTypeOrError(runtimeTypeName: $runtimeTypeName) {
                __typename
                ... on RegularRuntimeType {
                    name
                    displayName
                    isBuiltin
                }
                ... on RuntimeTypeNotFoundError {
                    runtimeTypeName
                }
            }
        }
        ... on PipelineNotFoundError {
            pipelineName
        }
    }
}
'''

ALL_RUNTIME_TYPES_QUERY = '''
fragment schemaTypeFragment on ConfigType {
  key
  ... on CompositeConfigType {
    fields {
      name
      configType {
        key
      }
    }
    recursiveConfigTypes {
      key
    }
  }
}
fragment runtimeTypeFragment on RuntimeType {
    key
    name
    displayName
    isNullable
    isList
    description
    inputSchemaType {
        ...schemaTypeFragment
    }
    outputSchemaType {
        ...schemaTypeFragment
    }
    innerTypes {
        key
    }
    ... on WrappingRuntimeType {
        ofType {
            key
        }
    }
}

{
 	pipelines {
    nodes {
      name
      runtimeTypes {
        ...runtimeTypeFragment
      }
    }
  }
}
'''


def test_dagster_type_query_works():
    result = execute_dagster_graphql(
        define_test_context(),
        RUNTIME_TYPE_QUERY,
        {'pipelineName': 'csv_hello_world', 'runtimeTypeName': 'PoorMansDataFrame'},
    )

    assert not result.errors
    assert result.data
    assert (
        result.data['pipelineOrError']['runtimeTypeOrError']['__typename'] == 'RegularRuntimeType'
    )
    assert result.data['pipelineOrError']['runtimeTypeOrError']['name'] == 'PoorMansDataFrame'


def test_dagster_type_query_with_container_context_ok():
    result = execute_dagster_graphql(
        define_test_snapshot_context(),
        RUNTIME_TYPE_QUERY,
        {'pipelineName': 'csv_hello_world', 'runtimeTypeName': 'PoorMansDataFrame'},
    )
    assert not result.errors
    assert result.data
    assert (
        result.data['pipelineOrError']['runtimeTypeOrError']['__typename'] == 'RegularRuntimeType'
    )
    assert result.data['pipelineOrError']['runtimeTypeOrError']['name'] == 'PoorMansDataFrame'


def test_dagster_type_builtin_query():
    result = execute_dagster_graphql(
        define_test_context(),
        RUNTIME_TYPE_QUERY,
        {'pipelineName': 'csv_hello_world', 'runtimeTypeName': 'Int'},
    )

    assert not result.errors
    assert result.data
    assert (
        result.data['pipelineOrError']['runtimeTypeOrError']['__typename'] == 'RegularRuntimeType'
    )
    assert result.data['pipelineOrError']['runtimeTypeOrError']['name'] == 'Int'
    assert result.data['pipelineOrError']['runtimeTypeOrError']['isBuiltin']


def test_dagster_type_or_error_pipeline_not_found():
    result = execute_dagster_graphql(
        define_test_context(),
        RUNTIME_TYPE_QUERY,
        {'pipelineName': 'nope', 'runtimeTypeName': 'nope'},
    )

    assert not result.errors
    assert result.data
    assert result.data['pipelineOrError']['__typename'] == 'PipelineNotFoundError'
    assert result.data['pipelineOrError']['pipelineName'] == 'nope'


def test_dagster_type_or_error_type_not_found():
    result = execute_dagster_graphql(
        define_test_context(),
        RUNTIME_TYPE_QUERY,
        {'pipelineName': 'csv_hello_world', 'runtimeTypeName': 'nope'},
    )

    assert not result.errors
    assert result.data
    assert (
        result.data['pipelineOrError']['runtimeTypeOrError']['__typename']
        == 'RuntimeTypeNotFoundError'
    )
    assert result.data['pipelineOrError']['runtimeTypeOrError']['runtimeTypeName'] == 'nope'


def test_dagster_type_or_error_query_with_container_context_not_found():
    result = execute_dagster_graphql(
        define_test_snapshot_context(),
        RUNTIME_TYPE_QUERY,
        {'pipelineName': 'csv_hello_world', 'runtimeTypeName': 'nope'},
    )

    assert not result.errors
    assert result.data
    assert (
        result.data['pipelineOrError']['runtimeTypeOrError']['__typename']
        == 'RuntimeTypeNotFoundError'
    )
    assert result.data['pipelineOrError']['runtimeTypeOrError']['runtimeTypeName'] == 'nope'


def test_smoke_test_dagster_type_system():
    result = execute_dagster_graphql(define_test_context(), ALL_RUNTIME_TYPES_QUERY)

    assert not result.errors
    assert result.data



from dagster_graphql.implementation.utils import ExecutionParams

from dagster import PipelineDefinition, PipelineRun, check
from dagster.core.execution.api import create_execution_plan
from dagster.core.execution.memoization import get_retry_steps_from_execution_plan
from dagster.core.instance import DagsterInstance
from dagster.core.storage.pipeline_run import PipelineRunStatus
from dagster.core.utils import make_new_run_id

from ..utils import UserFacingGraphQLError


def pipeline_run_from_execution_params(execution_params, step_keys_to_execute=None):
    check.inst_param(execution_params, 'execution_params', ExecutionParams)
    return PipelineRun(
        pipeline_name=execution_params.selector.name,
        run_id=execution_params.execution_metadata.run_id
        if execution_params.execution_metadata.run_id
        else make_new_run_id(),
        selector=execution_params.selector,
        environment_dict=execution_params.environment_dict,
        mode=execution_params.mode,
        step_keys_to_execute=step_keys_to_execute or execution_params.step_keys,
        tags=execution_params.execution_metadata.tags,
        root_run_id=(
            execution_params.execution_metadata.root_run_id or execution_params.previous_run_id
        ),
        parent_run_id=(
            execution_params.execution_metadata.parent_run_id or execution_params.previous_run_id
        ),
        previous_run_id=execution_params.previous_run_id,
        status=PipelineRunStatus.NOT_STARTED,
    )


def _create_pipeline_run(instance, pipeline_def, execution_params):
    check.inst_param(instance, 'instance', DagsterInstance)
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.inst_param(execution_params, 'execution_params', ExecutionParams)
    return pipeline_run_from_execution_params(
        execution_params, get_step_keys_to_execute(instance, pipeline_def, execution_params)
    )


def get_step_keys_to_execute(instance, pipeline_def, execution_params):
    check.inst_param(instance, 'instance', DagsterInstance)
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.inst_param(execution_params, 'execution_params', ExecutionParams)

    if not execution_params.step_keys and execution_params.previous_run_id:
        execution_plan = create_execution_plan(
            pipeline_def, execution_params.environment_dict, mode=execution_params.mode,
        )
        return get_retry_steps_from_execution_plan(
            instance, execution_plan, execution_params.previous_run_id
        )
    else:
        return execution_params.step_keys


def _check_start_pipeline_execution_errors(graphene_info, execution_params, execution_plan):
    if execution_params.step_keys:
        for step_key in execution_params.step_keys:
            if not execution_plan.has_step(step_key):
                raise UserFacingGraphQLError(
                    graphene_info.schema.type_named('InvalidStepError')(invalid_step_key=step_key)
                )



from __future__ import absolute_import

from dagster_graphql import dauphin
from dagster_graphql.implementation.utils import UserFacingGraphQLError, capture_dauphin_error

from dagster import PresetDefinition, check
from dagster.core.snap.config_types import ConfigSchemaSnapshot
from dagster.core.snap.mode import LoggerDefSnap, ModeDefSnap, ResourceDefSnap
from dagster.core.snap.pipeline_snapshot import PipelineIndex
from dagster.core.storage.pipeline_run import PipelineRunsFilter
from dagster.seven import lru_cache

from .config_types import DauphinConfigTypeField
from .runtime_types import to_dauphin_dagster_type
from .solids import DauphinSolidContainer, build_dauphin_solid_handles, build_dauphin_solids


class DauphinPipelineReference(dauphin.Interface):
    '''This interface supports the case where we can look up a pipeline successfully in the
    repository available to the DagsterInstance/graphql context, as well as the case where we know
    that a pipeline exists/existed thanks to materialized data such as logs and run metadata, but
    where we can't look the concrete pipeline up.'''

    class Meta(object):
        name = 'PipelineReference'

    name = dauphin.NonNull(dauphin.String)


class DauphinUnknownPipeline(dauphin.ObjectType):
    class Meta(object):
        name = 'UnknownPipeline'
        interfaces = (DauphinPipelineReference,)

    name = dauphin.NonNull(dauphin.String)


class DauphinIPipelineSnapshotMixin(object):
    # Mixin this class to implement IPipelineSnapshot
    #
    # Graphene has some strange properties that make it so that you cannot
    # implement ABCs nor use properties in an overridable way. So the way
    # the mixin works is that the target classes have to have a method
    # get_pipeline_index()
    #
    def get_pipeline_index(self):
        raise NotImplementedError()

    name = dauphin.NonNull(dauphin.String)
    description = dauphin.String()
    pipeline_snapshot_id = dauphin.NonNull(dauphin.String)
    runtime_types = dauphin.non_null_list('RuntimeType')
    runtime_type_or_error = dauphin.Field(
        dauphin.NonNull('RuntimeTypeOrError'),
        runtimeTypeName=dauphin.Argument(dauphin.NonNull(dauphin.String)),
    )
    solids = dauphin.non_null_list('Solid')
    modes = dauphin.non_null_list('Mode')
    solid_handles = dauphin.Field(
        dauphin.non_null_list('SolidHandle'), parentHandleID=dauphin.String()
    )
    solid_handle = dauphin.Field(
        'SolidHandle', handleID=dauphin.Argument(dauphin.NonNull(dauphin.String)),
    )
    tags = dauphin.non_null_list('PipelineTag')

    def resolve_pipeline_snapshot_id(self, _):
        return self.get_pipeline_index().pipeline_snapshot_id

    def resolve_name(self, _):
        return self.get_pipeline_index().name

    def resolve_description(self, _):
        return self.get_pipeline_index().description

    def resolve_runtime_types(self, _graphene_info):
        # TODO yuhan rename runtime_type in schema
        pipeline_index = self.get_pipeline_index()
        return sorted(
            list(
                map(
                    lambda dt: to_dauphin_dagster_type(pipeline_index.pipeline_snapshot, dt.key),
                    [t for t in pipeline_index.get_dagster_type_snaps() if t.name],
                )
            ),
            key=lambda dagster_type: dagster_type.name,
        )

    @capture_dauphin_error
    def resolve_runtime_type_or_error(self, _, **kwargs):
        type_name = kwargs['runtimeTypeName']

        pipeline_index = self.get_pipeline_index()

        if not pipeline_index.has_dagster_type_name(type_name):
            from .errors import DauphinRuntimeTypeNotFoundError

            raise UserFacingGraphQLError(
                DauphinRuntimeTypeNotFoundError(runtime_type_name=type_name)
            )

        return to_dauphin_dagster_type(
            pipeline_index.pipeline_snapshot,
            pipeline_index.get_dagster_type_from_name(type_name).key,
        )

    def resolve_solids(self, _graphene_info):
        pipeline_index = self.get_pipeline_index()
        return build_dauphin_solids(pipeline_index, pipeline_index.dep_structure_index)

    def resolve_modes(self, _):
        pipeline_snapshot = self.get_pipeline_index().pipeline_snapshot
        return [
            DauphinMode(pipeline_snapshot.config_schema_snapshot, mode_def_snap)
            for mode_def_snap in sorted(
                pipeline_snapshot.mode_def_snaps, key=lambda item: item.name
            )
        ]

    def resolve_solid_handle(self, _graphene_info, handleID):
        return _get_solid_handles(self.get_pipeline_index()).get(handleID)

    def resolve_solid_handles(self, _graphene_info, **kwargs):
        handles = _get_solid_handles(self.get_pipeline_index())
        parentHandleID = kwargs.get('parentHandleID')

        if parentHandleID == "":
            handles = {key: handle for key, handle in handles.items() if not handle.parent}
        elif parentHandleID is not None:
            handles = {
                key: handle
                for key, handle in handles.items()
                if handle.parent and handle.parent.handleID.to_string() == parentHandleID
            }

        return [handles[key] for key in sorted(handles)]

    def resolve_tags(self, graphene_info):
        return [
            graphene_info.schema.type_named('PipelineTag')(key=key, value=value)
            for key, value in self.get_pipeline_index().pipeline_snapshot.tags.items()
        ]


class DauphinIPipelineSnapshot(dauphin.Interface):
    class Meta(object):
        name = 'IPipelineSnapshot'

    name = dauphin.NonNull(dauphin.String)
    description = dauphin.String()
    pipeline_snapshot_id = dauphin.NonNull(dauphin.String)
    runtime_types = dauphin.non_null_list('RuntimeType')
    runtime_type_or_error = dauphin.Field(
        dauphin.NonNull('RuntimeTypeOrError'),
        runtimeTypeName=dauphin.Argument(dauphin.NonNull(dauphin.String)),
    )
    solids = dauphin.non_null_list('Solid')
    modes = dauphin.non_null_list('Mode')
    solid_handles = dauphin.Field(
        dauphin.non_null_list('SolidHandle'), parentHandleID=dauphin.String()
    )
    solid_handle = dauphin.Field(
        'SolidHandle', handleID=dauphin.Argument(dauphin.NonNull(dauphin.String)),
    )
    tags = dauphin.non_null_list('PipelineTag')


class DauphinPipeline(DauphinIPipelineSnapshotMixin, dauphin.ObjectType):
    class Meta(object):
        name = 'Pipeline'
        interfaces = (DauphinSolidContainer, DauphinPipelineReference, DauphinIPipelineSnapshot)

    presets = dauphin.non_null_list('PipelinePreset')
    runs = dauphin.non_null_list('PipelineRun')

    def __init__(self, pipeline_index, presets=None):
        self._pipeline_index = check.inst_param(pipeline_index, 'pipeline_index', PipelineIndex)
        self._presets = check.opt_list_param(presets, 'presets', of_type=PresetDefinition)

    def get_pipeline_index(self):
        return self._pipeline_index

    def resolve_presets(self, _graphene_info):
        check.list_param(self._presets, '_presets', of_type=PresetDefinition)
        return [
            DauphinPipelinePreset(preset, self._pipeline_index.name)
            for preset in sorted(self._presets, key=lambda item: item.name)
        ]

    def resolve_runs(self, graphene_info):
        return [
            graphene_info.schema.type_named('PipelineRun')(r)
            for r in graphene_info.context.instance.get_runs(
                filters=PipelineRunsFilter(pipeline_name=self.get_pipeline_index().name)
            )
        ]

    @staticmethod
    def from_pipeline_def(pipeline_definition):
        return DauphinPipeline(
            pipeline_definition.get_pipeline_index(), presets=pipeline_definition.get_presets()
        )


@lru_cache(maxsize=32)
def _get_solid_handles(pipeline_index):
    return {
        str(item.handleID): item
        for item in build_dauphin_solid_handles(pipeline_index, pipeline_index.dep_structure_index)
    }


class DauphinPipelineConnection(dauphin.ObjectType):
    class Meta(object):
        name = 'PipelineConnection'

    nodes = dauphin.non_null_list('Pipeline')


class DauphinResource(dauphin.ObjectType):
    class Meta(object):
        name = 'Resource'

    def __init__(self, config_schema_snapshot, resource_def_snap):
        self._config_schema_snapshot = check.inst_param(
            config_schema_snapshot, 'config_schema_snapshot', ConfigSchemaSnapshot
        )
        self._resource_dep_snap = check.inst_param(
            resource_def_snap, 'resource_def_snap', ResourceDefSnap
        )
        self.name = resource_def_snap.name
        self.description = resource_def_snap.description

    name = dauphin.NonNull(dauphin.String)
    description = dauphin.String()
    configField = dauphin.Field('ConfigTypeField')

    def resolve_configField(self, _):
        return (
            DauphinConfigTypeField(
                config_schema_snapshot=self._config_schema_snapshot,
                field_snap=self._resource_dep_snap.config_field_snap,
            )
            if self._resource_dep_snap.config_field_snap
            else None
        )


class DauphinLogger(dauphin.ObjectType):
    class Meta(object):
        name = 'Logger'

    def __init__(self, config_schema_snapshot, logger_def_snap):
        self._config_schema_snapshot = check.inst_param(
            config_schema_snapshot, 'config_schema_snapshot', ConfigSchemaSnapshot
        )
        self._logger_def_snap = check.inst_param(logger_def_snap, 'logger_def_snap', LoggerDefSnap)
        self.name = logger_def_snap.name
        self.description = logger_def_snap.description

    name = dauphin.NonNull(dauphin.String)
    description = dauphin.String()
    configField = dauphin.Field('ConfigTypeField')

    def resolve_configField(self, _):
        return (
            DauphinConfigTypeField(
                config_schema_snapshot=self._config_schema_snapshot,
                field_snap=self._logger_def_snap.config_field_snap,
            )
            if self._logger_def_snap.config_field_snap
            else None
        )


class DauphinMode(dauphin.ObjectType):
    def __init__(self, config_schema_snapshot, mode_def_snap):
        self._mode_def_snap = check.inst_param(mode_def_snap, 'mode_def_snap', ModeDefSnap)
        self._config_schema_snapshot = check.inst_param(
            config_schema_snapshot, 'config_schema_snapshot', ConfigSchemaSnapshot
        )

    class Meta(object):
        name = 'Mode'

    name = dauphin.NonNull(dauphin.String)
    description = dauphin.String()
    resources = dauphin.non_null_list('Resource')
    loggers = dauphin.non_null_list('Logger')

    def resolve_name(self, _graphene_info):
        return self._mode_def_snap.name

    def resolve_description(self, _graphene_info):
        return self._mode_def_snap.description

    def resolve_resources(self, _graphene_info):
        return [
            DauphinResource(self._config_schema_snapshot, resource_def_snap)
            for resource_def_snap in sorted(self._mode_def_snap.resource_def_snaps)
        ]

    def resolve_loggers(self, _graphene_info):
        return [
            DauphinLogger(self._config_schema_snapshot, logger_def_snap)
            for logger_def_snap in sorted(self._mode_def_snap.logger_def_snaps)
        ]


class DauphinMetadataItemDefinition(dauphin.ObjectType):
    class Meta(object):
        name = 'MetadataItemDefinition'

    key = dauphin.NonNull(dauphin.String)
    value = dauphin.NonNull(dauphin.String)


class DauphinPipelinePreset(dauphin.ObjectType):
    class Meta(object):
        name = 'PipelinePreset'

    name = dauphin.NonNull(dauphin.String)
    solidSubset = dauphin.List(dauphin.NonNull(dauphin.String))
    environmentConfigYaml = dauphin.NonNull(dauphin.String)
    mode = dauphin.NonNull(dauphin.String)

    def __init__(self, preset, pipeline_name):
        self._preset = check.inst_param(preset, 'preset', PresetDefinition)
        self._pipeline_name = check.str_param(pipeline_name, 'pipeline_name')

    def resolve_name(self, _graphene_info):
        return self._preset.name

    def resolve_solidSubset(self, _graphene_info):
        return self._preset.solid_subset

    def resolve_environmentConfigYaml(self, _graphene_info):
        yaml = self._preset.get_environment_yaml()
        return yaml if yaml else ''

    def resolve_mode(self, _graphene_info):
        return self._preset.mode


class DauphinPipelineSnapshot(DauphinIPipelineSnapshotMixin, dauphin.ObjectType):
    def __init__(self, pipeline_index):
        self._pipeline_index = check.inst_param(pipeline_index, 'pipeline_index', PipelineIndex)

    class Meta(object):
        name = 'PipelineSnapshot'
        interaces = (DauphinIPipelineSnapshot,)

    def get_pipeline_index(self):
        return self._pipeline_index


class DauphinPipelineSnapshotOrError(dauphin.Union):
    class Meta(object):
        name = 'PipelineSnapshotOrError'
        types = (
            'PipelineSnapshot',
            'PipelineSnapshotNotFoundError',
            'PipelineNotFoundError',
            'PythonError',
        )



from time import sleep

from dagster_graphql.test.utils import execute_dagster_graphql

from dagster import DagsterEventType
from dagster.core.instance import DagsterInstance
from dagster.core.utils import make_new_run_id

from .execution_queries import START_PIPELINE_EXECUTION_QUERY, START_PIPELINE_REEXECUTION_QUERY
from .setup import (
    define_test_context,
    define_test_subprocess_context,
    get_retry_multi_execution_params,
    retry_config,
)

NON_PERSISTENT_INTERMEDIATES_ERROR = (
    'Cannot perform reexecution with non persistent intermediates manager'
)


def step_did_not_run(logs, step_key):
    return not any(
        log['step']['key'] == step_key
        for log in logs
        if log['__typename']
        in ('ExecutionStepSuccessEvent', 'ExecutionStepSkippedEvent', 'ExecutionStepFailureEvent')
    )


def step_did_succeed(logs, step_key):
    return any(
        log['__typename'] == 'ExecutionStepSuccessEvent' and step_key == log['step']['key']
        for log in logs
    )


def step_did_skip(logs, step_key):
    return any(
        log['__typename'] == 'ExecutionStepSkippedEvent' and step_key == log['step']['key']
        for log in logs
    )


def step_did_fail(logs, step_key):
    return any(
        log['__typename'] == 'ExecutionStepFailureEvent' and step_key == log['step']['key']
        for log in logs
    )


def step_did_fail_in_records(records, step_key):
    return any(
        record.step_key == step_key
        and record.dagster_event.event_type_value == DagsterEventType.STEP_FAILURE.value
        for record in records
    )


def step_did_succeed_in_records(records, step_key):
    return any(
        record.step_key == step_key
        and record.dagster_event.event_type_value == DagsterEventType.STEP_SUCCESS.value
        for record in records
    )


def step_did_not_run_in_records(records, step_key):
    return not any(
        record.step_key == step_key
        and record.dagster_event.event_type_value
        in (
            DagsterEventType.STEP_SUCCESS.value,
            DagsterEventType.STEP_FAILURE.value,
            DagsterEventType.STEP_SKIPPED.value,
        )
        for record in records
    )


def test_retry_requires_intermediates():
    context = define_test_context()
    result = execute_dagster_graphql(
        context,
        START_PIPELINE_EXECUTION_QUERY,
        variables={
            'executionParams': {'mode': 'default', 'selector': {'name': 'eventually_successful'}}
        },
    )

    assert not result.errors
    assert result.data

    run_id = result.data['startPipelineExecution']['run']['runId']
    assert run_id
    logs = result.data['startPipelineExecution']['run']['logs']['nodes']
    assert isinstance(logs, list)

    assert step_did_succeed(logs, 'spawn.compute')
    assert step_did_fail(logs, 'fail.compute')
    assert step_did_skip(logs, 'fail_2.compute')
    assert step_did_skip(logs, 'fail_3.compute')
    assert step_did_skip(logs, 'reset.compute')

    retry_one = execute_dagster_graphql(
        context,
        START_PIPELINE_REEXECUTION_QUERY,
        variables={
            'executionParams': {
                'mode': 'default',
                'selector': {'name': 'eventually_successful'},
                'retryRunId': run_id,
                'executionMetadata': {'rootRunId': run_id, 'parentRunId': run_id,},
            }
        },
    )

    assert not retry_one.errors
    assert retry_one.data
    assert retry_one.data['startPipelineReexecution']['__typename'] == 'PythonError'
    assert (
        NON_PERSISTENT_INTERMEDIATES_ERROR in retry_one.data['startPipelineReexecution']['message']
    )


def test_retry_pipeline_execution():
    context = define_test_context(instance=DagsterInstance.local_temp())
    result = execute_dagster_graphql(
        context,
        START_PIPELINE_EXECUTION_QUERY,
        variables={
            'executionParams': {
                'mode': 'default',
                'selector': {'name': 'eventually_successful'},
                'environmentConfigData': retry_config(0),
            }
        },
    )

    run_id = result.data['startPipelineExecution']['run']['runId']
    logs = result.data['startPipelineExecution']['run']['logs']['nodes']
    assert step_did_succeed(logs, 'spawn.compute')
    assert step_did_fail(logs, 'fail.compute')
    assert step_did_skip(logs, 'fail_2.compute')
    assert step_did_skip(logs, 'fail_3.compute')
    assert step_did_skip(logs, 'reset.compute')

    retry_one = execute_dagster_graphql(
        context,
        START_PIPELINE_REEXECUTION_QUERY,
        variables={
            'executionParams': {
                'mode': 'default',
                'selector': {'name': 'eventually_successful'},
                'environmentConfigData': retry_config(1),
                'retryRunId': run_id,
                'executionMetadata': {'rootRunId': run_id, 'parentRunId': run_id,},
            }
        },
    )

    run_id = retry_one.data['startPipelineReexecution']['run']['runId']
    logs = retry_one.data['startPipelineReexecution']['run']['logs']['nodes']
    assert step_did_not_run(logs, 'spawn.compute')
    assert step_did_succeed(logs, 'fail.compute')
    assert step_did_fail(logs, 'fail_2.compute')
    assert step_did_skip(logs, 'fail_3.compute')
    assert step_did_skip(logs, 'reset.compute')

    retry_two = execute_dagster_graphql(
        context,
        START_PIPELINE_REEXECUTION_QUERY,
        variables={
            'executionParams': {
                'mode': 'default',
                'selector': {'name': 'eventually_successful'},
                'environmentConfigData': retry_config(2),
                'retryRunId': run_id,
                'executionMetadata': {'rootRunId': run_id, 'parentRunId': run_id,},
            }
        },
    )

    run_id = retry_two.data['startPipelineReexecution']['run']['runId']
    logs = retry_two.data['startPipelineReexecution']['run']['logs']['nodes']

    assert step_did_not_run(logs, 'spawn.compute')
    assert step_did_not_run(logs, 'fail.compute')
    assert step_did_succeed(logs, 'fail_2.compute')
    assert step_did_fail(logs, 'fail_3.compute')
    assert step_did_skip(logs, 'reset.compute')

    retry_three = execute_dagster_graphql(
        context,
        START_PIPELINE_REEXECUTION_QUERY,
        variables={
            'executionParams': {
                'mode': 'default',
                'selector': {'name': 'eventually_successful'},
                'environmentConfigData': retry_config(3),
                'retryRunId': run_id,
                'executionMetadata': {'rootRunId': run_id, 'parentRunId': run_id,},
            }
        },
    )

    run_id = retry_three.data['startPipelineReexecution']['run']['runId']
    logs = retry_three.data['startPipelineReexecution']['run']['logs']['nodes']

    assert step_did_not_run(logs, 'spawn.compute')
    assert step_did_not_run(logs, 'fail.compute')
    assert step_did_not_run(logs, 'fail_2.compute')
    assert step_did_succeed(logs, 'fail_3.compute')
    assert step_did_succeed(logs, 'reset.compute')


def test_retry_resource_pipeline():
    context = define_test_context(instance=DagsterInstance.local_temp())
    result = execute_dagster_graphql(
        context,
        START_PIPELINE_EXECUTION_QUERY,
        variables={
            'executionParams': {
                'mode': 'default',
                'selector': {'name': 'retry_resource_pipeline'},
                'environmentConfigData': {'storage': {'filesystem': {}}},
            }
        },
    )

    run_id = result.data['startPipelineExecution']['run']['runId']
    logs = result.data['startPipelineExecution']['run']['logs']['nodes']
    assert step_did_succeed(logs, 'start.compute')
    assert step_did_fail(logs, 'will_fail.compute')

    retry_one = execute_dagster_graphql(
        context,
        START_PIPELINE_REEXECUTION_QUERY,
        variables={
            'executionParams': {
                'mode': 'default',
                'selector': {'name': 'retry_resource_pipeline'},
                'environmentConfigData': {'storage': {'filesystem': {}}},
                'retryRunId': run_id,
                'executionMetadata': {'rootRunId': run_id, 'parentRunId': run_id,},
            }
        },
    )
    run_id = retry_one.data['startPipelineReexecution']['run']['runId']
    logs = retry_one.data['startPipelineReexecution']['run']['logs']['nodes']
    assert step_did_not_run(logs, 'start.compute')
    assert step_did_fail(logs, 'will_fail.compute')


def test_retry_multi_output():
    context = define_test_context(instance=DagsterInstance.local_temp())
    result = execute_dagster_graphql(
        context,
        START_PIPELINE_EXECUTION_QUERY,
        variables={'executionParams': get_retry_multi_execution_params(should_fail=True)},
    )

    print(result.data)
    run_id = result.data['startPipelineExecution']['run']['runId']
    logs = result.data['startPipelineExecution']['run']['logs']['nodes']
    assert step_did_succeed(logs, 'multi.compute')
    assert step_did_skip(logs, 'child_multi_skip.compute')
    assert step_did_fail(logs, 'can_fail.compute')
    assert step_did_skip(logs, 'child_fail.compute')
    assert step_did_skip(logs, 'child_skip.compute')
    assert step_did_skip(logs, 'grandchild_fail.compute')

    retry_one = execute_dagster_graphql(
        context,
        START_PIPELINE_REEXECUTION_QUERY,
        variables={
            'executionParams': get_retry_multi_execution_params(should_fail=True, retry_id=run_id)
        },
    )

    run_id = retry_one.data['startPipelineReexecution']['run']['runId']
    logs = retry_one.data['startPipelineReexecution']['run']['logs']['nodes']
    assert step_did_not_run(logs, 'multi.compute')
    assert step_did_not_run(logs, 'child_multi_skip.compute')
    assert step_did_fail(logs, 'can_fail.compute')
    assert step_did_skip(logs, 'child_fail.compute')
    assert step_did_skip(logs, 'child_skip.compute')
    assert step_did_skip(logs, 'grandchild_fail.compute')

    retry_two = execute_dagster_graphql(
        context,
        START_PIPELINE_REEXECUTION_QUERY,
        variables={
            'executionParams': get_retry_multi_execution_params(should_fail=False, retry_id=run_id)
        },
    )

    run_id = retry_two.data['startPipelineReexecution']['run']['runId']
    logs = retry_two.data['startPipelineReexecution']['run']['logs']['nodes']
    assert step_did_not_run(logs, 'multi.compute')
    assert step_did_not_run(logs, 'child_multi_skip.compute')
    assert step_did_succeed(logs, 'can_fail.compute')
    assert step_did_succeed(logs, 'child_fail.compute')
    assert step_did_skip(logs, 'child_skip.compute')
    assert step_did_succeed(logs, 'grandchild_fail.compute')


def test_retry_early_terminate():
    instance = DagsterInstance.local_temp()
    context = define_test_subprocess_context(instance=instance)
    run_id = make_new_run_id()
    execute_dagster_graphql(
        context,
        START_PIPELINE_EXECUTION_QUERY,
        variables={
            'executionParams': {
                'mode': 'default',
                'selector': {'name': 'retry_multi_input_early_terminate_pipeline'},
                'environmentConfigData': {
                    'solids': {
                        'get_input_one': {'config': {'wait_to_terminate': True}},
                        'get_input_two': {'config': {'wait_to_terminate': True}},
                    },
                    'storage': {'filesystem': {}},
                },
                'executionMetadata': {'runId': run_id},
            }
        },
    )
    # Wait until the first step succeeded
    while instance.get_run_stats(run_id).steps_succeeded < 1:
        sleep(0.1)
    # Terminate the current pipeline run at the second step
    context.execution_manager.terminate(run_id)

    records = instance.all_logs(run_id)

    # The first step should succeed, the second should fail or not start,
    # and the following steps should not appear in records
    assert step_did_succeed_in_records(records, 'return_one.compute')
    assert any(
        [
            step_did_fail_in_records(records, 'get_input_one.compute'),
            step_did_not_run_in_records(records, 'get_input_one.compute'),
        ]
    )
    assert step_did_not_run_in_records(records, 'get_input_two.compute')
    assert step_did_not_run_in_records(records, 'sum_inputs.compute')

    # Start retry
    new_run_id = make_new_run_id()

    execute_dagster_graphql(
        context,
        START_PIPELINE_REEXECUTION_QUERY,
        variables={
            'executionParams': {
                'mode': 'default',
                'selector': {'name': 'retry_multi_input_early_terminate_pipeline'},
                'environmentConfigData': {
                    'solids': {
                        'get_input_one': {'config': {'wait_to_terminate': False}},
                        'get_input_two': {'config': {'wait_to_terminate': False}},
                    },
                    'storage': {'filesystem': {}},
                },
                'retryRunId': run_id,
                'executionMetadata': {
                    'runId': new_run_id,
                    'rootRunId': run_id,
                    'parentRunId': run_id,
                },
            }
        },
    )
    # Wait until the run is finished
    while context.execution_manager.is_process_running(new_run_id):
        pass

    retry_records = instance.all_logs(new_run_id)
    # The first step should not run and the other three steps should succeed in retry
    assert step_did_not_run_in_records(retry_records, 'return_one.compute')
    assert step_did_succeed_in_records(retry_records, 'get_input_one.compute')
    assert step_did_succeed_in_records(retry_records, 'get_input_two.compute')
    assert step_did_succeed_in_records(retry_records, 'sum_inputs.compute')



from __future__ import absolute_import

import sys
import time

from graphql.execution.base import ResolveInfo

from dagster import check
from dagster.core.definitions.schedule import ScheduleExecutionContext
from dagster.core.errors import ScheduleExecutionError, user_code_error_boundary
from dagster.core.scheduler import ScheduleTickStatus
from dagster.core.scheduler.scheduler import ScheduleTickData
from dagster.utils.error import serializable_error_info_from_exc_info

from ..fetch_pipelines import get_pipeline_def_from_selector
from ..fetch_schedules import execution_params_for_schedule, get_dagster_schedule_def
from ..utils import capture_dauphin_error
from .launch_execution import _launch_pipeline_execution
from .start_execution import _start_pipeline_execution


@capture_dauphin_error
def start_scheduled_execution(graphene_info, schedule_name):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.str_param(schedule_name, 'schedule_name')

    tick = None
    try:
        repository = graphene_info.context.get_repository()
        schedule_context = ScheduleExecutionContext(graphene_info.context.instance, repository)
        schedule_def = get_dagster_schedule_def(graphene_info, schedule_name)
        tick = graphene_info.context.instance.create_schedule_tick(
            repository,
            ScheduleTickData(
                schedule_name=schedule_def.name,
                cron_schedule=schedule_def.cron_schedule,
                timestamp=time.time(),
                status=ScheduleTickStatus.STARTED,
            ),
        )

        # Run should_execute and halt if it returns False
        with user_code_error_boundary(
            ScheduleExecutionError,
            lambda: 'Error occurred during the execution should_execute for schedule '
            '{schedule_name}'.format(schedule_name=schedule_def.name),
        ):
            should_execute = schedule_def.should_execute(schedule_context)

        if not should_execute:
            # Update tick
            tick = tick.with_status(ScheduleTickStatus.SKIPPED)
            graphene_info.context.instance.update_schedule_tick(repository, tick)
            # Return skipped specific gql response
            return graphene_info.schema.type_named('ScheduledExecutionBlocked')(
                message='Schedule {schedule_name} did not run because the should_execute did not return'
                ' True'.format(schedule_name=schedule_name)
            )

        pipeline_def = get_pipeline_def_from_selector(graphene_info, schedule_def.selector)
        execution_params = execution_params_for_schedule(
            graphene_info, schedule_def=schedule_def, pipeline_def=pipeline_def
        )

        # Launch run if run launcher is defined
        run_launcher = graphene_info.context.instance.run_launcher
        if run_launcher:
            result = _launch_pipeline_execution(graphene_info, execution_params)
        else:
            result = _start_pipeline_execution(graphene_info, execution_params)

        run = result.run
        graphene_info.context.instance.update_schedule_tick(
            repository, tick.with_status(ScheduleTickStatus.SUCCESS, run_id=run.run_id),
        )

        return result

    except Exception as exc:  # pylint: disable=broad-except
        error_data = serializable_error_info_from_exc_info(sys.exc_info())

        if tick:
            graphene_info.context.instance.update_schedule_tick(
                repository, tick.with_status(ScheduleTickStatus.FAILURE, error=error_data),
            )

        raise exc



from contextlib import contextmanager

import pytest
from dagster_graphql.implementation.fetch_pipelines import (
    _get_dauphin_pipeline_snapshot_from_instance,
)
from dagster_graphql.implementation.utils import UserFacingGraphQLError
from dagster_graphql.test.utils import execute_dagster_graphql

from dagster import execute_pipeline
from dagster.core.instance import DagsterInstance
from dagster.seven import json, mock

from .setup import define_test_context, noop_pipeline

SNAPSHOT_QUERY = '''
query PipelineSnapshotQuery($snapshotId: String!) {
    pipelineSnapshot(snapshotId: $snapshotId) {
        __typename
        name
        pipelineSnapshotId
        description
        runtimeTypes { key }
        solids { name }
        modes { name }
        solidHandles { handleID }
        tags { key value }
    }
}
'''

SNAPSHOT_OR_ERROR_QUERY_BY_SNAPSHOT_ID = '''
query PipelineSnapshotQueryBySnapshotID($snapshotId: String!) {
    pipelineSnapshotOrError(snapshotId: $snapshotId) {
        __typename
        ... on PipelineSnapshot {
            name
            pipelineSnapshotId
            description
            runtimeTypes { key }
            solids { name }
            modes { name }
            solidHandles { handleID }
            tags { key value }
        }
        ... on PipelineSnapshotNotFoundError {
            snapshotId
        }
    }
}
'''

SNAPSHOT_OR_ERROR_QUERY_BY_PIPELINE_NAME = '''
query PipelineSnapshotQueryByActivePipelineName($activePipelineName: String!) {
    pipelineSnapshotOrError(activePipelineName: $activePipelineName) {
        __typename
        ... on PipelineSnapshot {
            name
            pipelineSnapshotId
            description
            runtimeTypes { key }
            solids { name }
            modes { name }
            solidHandles { handleID }
            tags { key value }
        }
        ... on PipelineSnapshotNotFoundError {
            snapshotId
        }
    }
}
'''

# makes snapshot tests much easier to debug
def pretty_dump(data):
    return json.dumps(data, indent=2, separators=(',', ': '))


@contextmanager
def create_ephemeral_instance():
    yield DagsterInstance.ephemeral()


@contextmanager
def create_local_temp_instance():
    yield DagsterInstance.local_temp()


class TestPipelineSnapshotGraphQL:
    __test__ = True

    @pytest.fixture(name='instance', params=[create_ephemeral_instance, create_local_temp_instance])
    def get_instance(self, request):
        with request.param() as s:
            yield s

    def test_fetch_snapshot_success(self, instance, snapshot):
        result = execute_pipeline(noop_pipeline, instance=instance)
        assert result.success
        run = instance.get_run_by_id(result.run_id)
        assert run.pipeline_snapshot_id

        result = execute_dagster_graphql(
            define_test_context(instance), SNAPSHOT_QUERY, {'snapshotId': run.pipeline_snapshot_id}
        )

        assert not result.errors
        assert result.data
        assert result.data['pipelineSnapshot']['__typename'] == 'PipelineSnapshot'
        snapshot.assert_match(pretty_dump(result.data))

    def test_fetch_snapshot_or_error_by_snapshot_id_success(self, instance, snapshot):
        result = execute_pipeline(noop_pipeline, instance=instance)
        assert result.success
        run = instance.get_run_by_id(result.run_id)
        assert run.pipeline_snapshot_id

        result = execute_dagster_graphql(
            define_test_context(instance),
            SNAPSHOT_OR_ERROR_QUERY_BY_SNAPSHOT_ID,
            {'snapshotId': run.pipeline_snapshot_id},
        )

        assert not result.errors
        assert result.data
        assert result.data['pipelineSnapshotOrError']['__typename'] == 'PipelineSnapshot'

        snapshot.assert_match(pretty_dump(result.data))

    def test_fetch_snapshot_or_error_by_snapshot_id_snapshot_not_found(self, instance, snapshot):
        result = execute_dagster_graphql(
            define_test_context(instance),
            SNAPSHOT_OR_ERROR_QUERY_BY_SNAPSHOT_ID,
            {'snapshotId': 'notthere'},
        )

        assert not result.errors
        assert result.data
        assert (
            result.data['pipelineSnapshotOrError']['__typename'] == 'PipelineSnapshotNotFoundError'
        )
        assert result.data['pipelineSnapshotOrError']['snapshotId'] == 'notthere'
        snapshot.assert_match(pretty_dump(result.data))

    def test_fetch_snapshot_or_error_by_active_pipeline_name_success(self, instance, snapshot):
        result = execute_dagster_graphql(
            define_test_context(instance),
            SNAPSHOT_OR_ERROR_QUERY_BY_PIPELINE_NAME,
            {'activePipelineName': 'csv_hello_world'},
        )

        assert not result.errors
        assert result.data
        assert result.data['pipelineSnapshotOrError']['__typename'] == 'PipelineSnapshot'
        assert result.data['pipelineSnapshotOrError']['name'] == 'csv_hello_world'

        snapshot.assert_match(pretty_dump(result.data))

    def test_fetch_snapshot_or_error_by_active_pipeline_name_not_found(self, instance, snapshot):
        result = execute_dagster_graphql(
            define_test_context(instance),
            SNAPSHOT_OR_ERROR_QUERY_BY_PIPELINE_NAME,
            {'activePipelineName': 'jkdjfkdj'},
        )

        assert not result.errors
        assert result.data
        assert result.data['pipelineSnapshotOrError']['__typename'] == 'PipelineNotFoundError'

        snapshot.assert_match(pretty_dump(result.data))


def test_temporary_error_or_deletion_after_instance_check():
    instance = mock.MagicMock()

    instance.has_pipeline_snapshot.return_value = True
    instance.get_pipeline_snapshot.return_value = None

    with pytest.raises(UserFacingGraphQLError):
        _get_dauphin_pipeline_snapshot_from_instance(instance, 'kjdkfjd')



from dagster_graphql.client.util import execution_params_from_pipeline_run
from dagster_graphql.implementation.execution.utils import pipeline_run_from_execution_params
from dagster_graphql.schema.roots import execution_params_from_graphql

from dagster.core.definitions.pipeline import ExecutionSelector
from dagster.core.storage.pipeline_run import PipelineRun, PipelineRunStatus


def test_roundtrip_run():
    run_with_snapshot = PipelineRun(
        pipeline_name='pipey_mcpipeface',
        run_id='8675309',
        environment_dict={'good': True},
        mode='default',
        selector=ExecutionSelector('pipey_mcpipeface'),
        step_keys_to_execute=['step_1', 'step_2', 'step_3'],
        tags={'tag_it': 'bag_it'},
        status=PipelineRunStatus.NOT_STARTED,
        root_run_id='previousID',
        parent_run_id='previousID',
        previous_run_id='previousID',
        pipeline_snapshot_id='pipey_mcpipeface_snapshot_id',
        execution_plan_snapshot_id='mcexecutionplanface_snapshot_id',
    )
    for field in run_with_snapshot:
        # ensure we have a test value to round trip for each field
        assert field

    # The invariant that all the execution parameter structs
    # pipeline run can be constructed from each other is no longer
    # true. Clients of the GraphQL API cannot know the value of the
    # pipeline_snapshot_id prior to execution, because it is
    # constructed on the server. Hence these roundtrip tests
    # do not include snapshot_id

    run = run_with_snapshot._replace(pipeline_snapshot_id=None, execution_plan_snapshot_id=None)

    exec_params = execution_params_from_pipeline_run(run)
    assert run == pipeline_run_from_execution_params(exec_params)

    exec_params_gql = execution_params_from_graphql(exec_params.to_graphql_input())
    assert exec_params_gql == exec_params
    assert run == pipeline_run_from_execution_params(exec_params_gql)

    empty_run = PipelineRun(pipeline_name='foo', run_id='bar', mode='default')
    exec_params = execution_params_from_pipeline_run(empty_run)
    assert empty_run == pipeline_run_from_execution_params(exec_params)

    exec_params_gql = execution_params_from_graphql(exec_params.to_graphql_input())
    assert exec_params_gql == exec_params
    assert empty_run == pipeline_run_from_execution_params(exec_params_gql)



from dagster_graphql.implementation.context import DagsterGraphQLContext
from dagster_graphql.implementation.pipeline_execution_manager import (
    SubprocessExecutionManager,
    SynchronousExecutionManager,
)
from dagster_graphql.schema import create_schema
from graphql import graphql

from dagster import ExecutionTargetHandle
from dagster.core.instance import DagsterInstance


def execute_dagster_graphql(context, query, variables=None):
    result = graphql(
        create_schema(),
        query,
        context_value=context,
        variable_values=variables,
        # executor=GeventObservableExecutor(),
        allow_subscriptions=True,
        return_promise=False,
    )

    # has to check attr because in subscription case it returns AnonymousObservable
    if hasattr(result, 'errors') and result.errors:
        first_error = result.errors[0]
        if hasattr(first_error, 'original_error') and first_error.original_error:
            raise result.errors[0].original_error

        raise result.errors[0]

    return result


def define_context(repo_fn, instance=None):
    return DagsterGraphQLContext(
        handle=ExecutionTargetHandle.for_repo_fn(repo_fn),
        instance=instance or DagsterInstance.ephemeral(),
        execution_manager=SynchronousExecutionManager(),
    )


def define_subprocess_context(repo_fn, instance):
    return DagsterGraphQLContext(
        handle=ExecutionTargetHandle.for_repo_fn(repo_fn),
        instance=instance,
        execution_manager=SubprocessExecutionManager(instance),
    )


def define_context_for_file(python_file, fn_name, instance=None):
    return DagsterGraphQLContext(
        handle=ExecutionTargetHandle.for_repo_python_file(python_file, fn_name),
        instance=instance or DagsterInstance.ephemeral(),
        execution_manager=SynchronousExecutionManager(),
    )


def define_subprocess_context_for_file(python_file, fn_name, instance=None):
    return DagsterGraphQLContext(
        handle=ExecutionTargetHandle.for_repo_python_file(python_file, fn_name),
        instance=instance or DagsterInstance.ephemeral(),
        execution_manager=SubprocessExecutionManager(instance),
    )


def define_context_for_repository_yaml(path, instance=None):
    return DagsterGraphQLContext(
        handle=ExecutionTargetHandle.for_repo_yaml(path),
        instance=instance or DagsterInstance.ephemeral(),
        execution_manager=SynchronousExecutionManager(),
    )



from __future__ import absolute_import

from dagster_graphql.schema.errors import DauphinPipelineConfigValidationInvalid
from graphql.execution.base import ResolveInfo

from dagster import DagsterInvalidConfigError, check
from dagster.config.validate import validate_config
from dagster.core.definitions.environment_schema import create_environment_schema
from dagster.core.events import EngineEventData
from dagster.core.execution.api import create_execution_plan
from dagster.core.instance import InstanceCreateRunArgs
from dagster.core.snap.execution_plan_snapshot import snapshot_from_execution_plan
from dagster.core.storage.pipeline_run import PipelineRunStatus
from dagster.core.utils import make_new_run_id
from dagster.utils import merge_dicts
from dagster.utils.error import SerializableErrorInfo

from ..fetch_pipelines import get_pipeline_def_from_selector
from ..fetch_runs import get_validated_config
from ..utils import ExecutionMetadata, ExecutionParams, capture_dauphin_error
from .utils import (
    _check_start_pipeline_execution_errors,
    _create_pipeline_run,
    get_step_keys_to_execute,
)


@capture_dauphin_error
def start_pipeline_reexecution(graphene_info, execution_params):
    return _start_pipeline_execution(graphene_info, execution_params, is_reexecuted=True)


@capture_dauphin_error
def start_pipeline_execution(graphene_info, execution_params):
    '''This indirection done on purpose to make the logic in the function
    below re-usable. The parent function is wrapped in @capture_dauphin_error, which makes it
    difficult to do exception handling.
    '''
    return _start_pipeline_execution(graphene_info, execution_params)


def _start_pipeline_execution(graphene_info, execution_params, is_reexecuted=False):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(execution_params, 'execution_params', ExecutionParams)

    if is_reexecuted:
        # required fields for re-execution
        execution_metadata = check.inst_param(
            execution_params.execution_metadata, 'execution_metadata', ExecutionMetadata
        )
        check.str_param(execution_metadata.root_run_id, 'root_run_id')
        check.str_param(execution_metadata.parent_run_id, 'parent_run_id')

    error_type = (
        'StartPipelineExecutionDisabledError'
        if not is_reexecuted
        else 'DauphinStartPipelineReexecutionDisabledError'
    )
    success_type = (
        'StartPipelineExecutionSuccess' if not is_reexecuted else 'StartPipelineReexecutionSuccess'
    )

    instance = graphene_info.context.instance

    execution_manager_settings = instance.dagit_settings.get('execution_manager')
    if execution_manager_settings and execution_manager_settings.get('disabled'):
        return graphene_info.schema.type_named(error_type)()

    pipeline_def = get_pipeline_def_from_selector(graphene_info, execution_params.selector)

    get_validated_config(
        pipeline_def,
        environment_dict=execution_params.environment_dict,
        mode=execution_params.mode,
    )

    execution_plan = create_execution_plan(
        pipeline_def, execution_params.environment_dict, mode=execution_params.mode,
    )

    _check_start_pipeline_execution_errors(graphene_info, execution_params, execution_plan)

    if execution_params.execution_metadata.run_id:
        # If a run_id is provided, use the old get_or_create_run machinery
        run = instance.get_or_create_run(
            _create_pipeline_run(instance, pipeline_def, execution_params)
        )
    else:
        # Otherwise we know we are creating a new run, and we can
        # use the new machinery that persists a pipeline snapshot
        # with the run.

        run = instance.create_run_with_snapshot(
            InstanceCreateRunArgs(
                pipeline_snapshot=pipeline_def.get_pipeline_snapshot(),
                execution_plan_snapshot=snapshot_from_execution_plan(
                    execution_plan, pipeline_def.get_pipeline_snapshot_id()
                ),
                run_id=execution_params.execution_metadata.run_id
                if execution_params.execution_metadata.run_id
                else make_new_run_id(),
                selector=execution_params.selector,
                environment_dict=execution_params.environment_dict,
                mode=execution_params.mode,
                step_keys_to_execute=get_step_keys_to_execute(
                    instance, pipeline_def, execution_params
                )
                or execution_params.step_keys,
                tags=merge_dicts(pipeline_def.tags, execution_params.execution_metadata.tags),
                status=PipelineRunStatus.NOT_STARTED,
                root_run_id=(
                    execution_params.execution_metadata.root_run_id
                    or execution_params.previous_run_id
                ),
                parent_run_id=(
                    execution_params.execution_metadata.parent_run_id
                    or execution_params.previous_run_id
                ),
            )
        )

    graphene_info.context.execution_manager.execute_pipeline(
        graphene_info.context.get_handle(), pipeline_def, run, instance=instance,
    )

    return graphene_info.schema.type_named(success_type)(
        run=graphene_info.schema.type_named('PipelineRun')(run)
    )


@capture_dauphin_error
def start_pipeline_execution_for_created_run(graphene_info, run_id):
    '''This indirection is done on purpose to make the logic in the function
    below re-usable. The parent function is wrapped in @capture_dauphin_error, which makes it
    difficult to do exception handling.
    '''
    return _start_pipeline_execution_for_created_run(graphene_info, run_id)


def _start_pipeline_execution_for_created_run(graphene_info, run_id):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)

    instance = graphene_info.context.instance

    execution_manager_settings = instance.dagit_settings.get('execution_manager')
    if execution_manager_settings and execution_manager_settings.get('disabled'):
        return graphene_info.schema.type_named('StartPipelineExecutionDisabledError')()

    pipeline_run = instance.get_run_by_id(run_id)
    if not pipeline_run:
        return graphene_info.schema.type_named('PipelineRunNotFoundError')(run_id)

    pipeline_def = get_pipeline_def_from_selector(graphene_info, pipeline_run.selector)

    environment_schema = create_environment_schema(pipeline_def, pipeline_run.mode)
    validated_config = validate_config(
        environment_schema.environment_type, pipeline_run.environment_dict
    )
    if not validated_config.success:
        # If the config is invalid, we construct a DagsterInvalidConfigError exception and
        # insert it into the event log. We also return a PipelineConfigValidationInvalid user facing
        # graphql error.

        # We currently re-use the engine events machinery to add the error to the event log, but
        # may need to create a new event type and instance method to handle these erros.
        invalid_config_exception = DagsterInvalidConfigError(
            'Error in config for pipeline {}'.format(pipeline_def.name),
            validated_config.errors,
            pipeline_run.environment_dict,
        )

        instance.report_engine_event(
            str(invalid_config_exception.message),
            pipeline_run,
            EngineEventData.engine_error(
                SerializableErrorInfo(
                    invalid_config_exception.message,
                    [],
                    DagsterInvalidConfigError.__class__.__name__,
                    None,
                )
            ),
        )

        # TODO: also insert a pipeline init failure event
        # https://github.com/dagster-io/dagster/issues/2385

        return DauphinPipelineConfigValidationInvalid.for_validation_errors(
            pipeline_def, validated_config.errors
        )

    create_execution_plan(
        pipeline_def,
        pipeline_run.environment_dict,
        mode=pipeline_run.mode,
        step_keys_to_execute=pipeline_run.step_keys_to_execute,
    )

    graphene_info.context.execution_manager.execute_pipeline(
        graphene_info.context.get_handle(), pipeline_def, pipeline_run, instance=instance,
    )

    return graphene_info.schema.type_named('StartPipelineExecutionSuccess')(
        run=graphene_info.schema.type_named('PipelineRun')(pipeline_run)
    )



from graphql.execution.base import ResolveInfo

from dagster import check
from dagster.core.definitions.pipeline import ExecutionSelector

from .fetch_pipelines import get_pipeline_def_from_selector
from .utils import capture_dauphin_error


@capture_dauphin_error
def get_partition_sets_or_error(graphene_info, pipeline_name):
    return graphene_info.schema.type_named('PartitionSets')(
        results=_get_partition_sets(graphene_info, pipeline_name)
    )


def _get_partition_sets(graphene_info, pipeline_name):
    partition_sets = graphene_info.context.get_all_partition_sets()

    if pipeline_name:
        pipeline_def = get_pipeline_def_from_selector(
            graphene_info, ExecutionSelector(pipeline_name)
        )
        matching_partition_sets = filter(
            lambda partition_set: partition_set.pipeline_name == pipeline_def.name, partition_sets
        )
    else:
        matching_partition_sets = partition_sets

    return [
        graphene_info.schema.type_named('PartitionSet')(partition_set)
        for partition_set in matching_partition_sets
    ]


@capture_dauphin_error
def get_partition_set(graphene_info, partition_set_name):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    partition_set = graphene_info.context.get_partition_set(partition_set_name)
    return graphene_info.schema.type_named('PartitionSet')(partition_set)



import heapq
import json
import os

import yaml
from dagster_graphql import dauphin
from dagster_graphql.implementation.fetch_schedules import (
    get_dagster_schedule_def,
    get_schedule_attempt_filenames,
)
from dagster_graphql.schema.errors import (
    DauphinScheduleNotFoundError,
    DauphinSchedulerNotDefinedError,
)

from dagster import check
from dagster.core.definitions import ScheduleDefinition, ScheduleExecutionContext
from dagster.core.definitions.partition import PartitionScheduleDefinition
from dagster.core.errors import ScheduleExecutionError, user_code_error_boundary
from dagster.core.scheduler import Schedule, ScheduleTickStatus
from dagster.core.scheduler.scheduler import ScheduleTickStatsSnapshot
from dagster.core.storage.pipeline_run import PipelineRunsFilter


class DauphinScheduleStatus(dauphin.Enum):
    class Meta(object):
        name = 'ScheduleStatus'

    RUNNING = 'RUNNING'
    STOPPED = 'STOPPED'
    ENDED = 'ENDED'


class DauphinSchedulerOrError(dauphin.Union):
    class Meta(object):
        name = 'SchedulerOrError'
        types = ('Scheduler', DauphinSchedulerNotDefinedError, 'PythonError')


class DauphinScheduleOrError(dauphin.Union):
    class Meta(object):
        name = 'ScheduleOrError'
        types = ('RunningSchedule', DauphinScheduleNotFoundError, 'PythonError')


class DauphinScheduleDefinition(dauphin.ObjectType):
    class Meta(object):
        name = 'ScheduleDefinition'

    name = dauphin.NonNull(dauphin.String)
    cron_schedule = dauphin.NonNull(dauphin.String)

    pipeline_name = dauphin.NonNull(dauphin.String)
    solid_subset = dauphin.List(dauphin.String)
    mode = dauphin.NonNull(dauphin.String)
    environment_config_yaml = dauphin.Field(dauphin.String)
    partition_set = dauphin.Field('PartitionSet')

    def resolve_environment_config_yaml(self, _graphene_info):
        schedule_def = self._schedule_def
        try:
            with user_code_error_boundary(
                ScheduleExecutionError,
                lambda: 'Error occurred during the execution of environment_dict_fn for schedule '
                '{schedule_name}'.format(schedule_name=schedule_def.name),
            ):
                environment_config = schedule_def.get_environment_dict(self._schedule_context)
        except ScheduleExecutionError:
            return None

        environment_config_yaml = yaml.dump(environment_config, default_flow_style=False)
        return environment_config_yaml if environment_config_yaml else ''

    def resolve_partition_set(self, graphene_info):
        if isinstance(self._schedule_def, PartitionScheduleDefinition):
            return graphene_info.schema.type_named('PartitionSet')(
                self._schedule_def.get_partition_set()
            )

        return None

    def __init__(self, graphene_info, schedule_def):
        self._schedule_def = check.inst_param(schedule_def, 'schedule_def', ScheduleDefinition)
        self._schedule_context = ScheduleExecutionContext(
            graphene_info.context.instance, graphene_info.context.get_repository()
        )
        self._schedule_def = check.inst_param(schedule_def, 'schedule_def', ScheduleDefinition)

        super(DauphinScheduleDefinition, self).__init__(
            name=schedule_def.name,
            cron_schedule=schedule_def.cron_schedule,
            pipeline_name=schedule_def.selector.name,
            solid_subset=schedule_def.selector.solid_subset,
            mode=schedule_def.mode,
        )


# TODO: Delete in 0.8.0 release
# https://github.com/dagster-io/dagster/issues/2288
class DauphinScheduleAttemptStatus(dauphin.Enum):
    class Meta(object):
        name = 'ScheduleAttemptStatus'

    SUCCESS = 'SUCCESS'
    ERROR = 'ERROR'
    SKIPPED = 'SKIPPED'


# TODO: Delete in 0.8.0 release
# https://github.com/dagster-io/dagster/issues/228
class DauphinScheduleAttempt(dauphin.ObjectType):
    class Meta(object):
        name = 'ScheduleAttempt'

    time = dauphin.NonNull(dauphin.Float)
    json_result = dauphin.NonNull(dauphin.String)
    status = dauphin.NonNull('ScheduleAttemptStatus')
    run = dauphin.Field('PipelineRun')


DauphinScheduleTickStatus = dauphin.Enum.from_enum(ScheduleTickStatus)


class DauphinScheduleTick(dauphin.ObjectType):
    class Meta(object):
        name = 'ScheduleTick'

    tick_id = dauphin.NonNull(dauphin.String)
    status = dauphin.NonNull('ScheduleTickStatus')
    timestamp = dauphin.NonNull(dauphin.Float)
    tick_specific_data = dauphin.Field('ScheduleTickSpecificData')


class DauphinScheduleTickSuccessData(dauphin.ObjectType):
    class Meta(object):
        name = 'ScheduleTickSuccessData'

    run = dauphin.Field('PipelineRun')


class DauphinScheduleTickFailureData(dauphin.ObjectType):
    class Meta(object):
        name = 'ScheduleTickFailureData'

    error = dauphin.NonNull('PythonError')


class DauphinScheduleTickSpecificData(dauphin.Union):
    class Meta(object):
        name = 'ScheduleTickSpecificData'
        types = (
            DauphinScheduleTickSuccessData,
            DauphinScheduleTickFailureData,
        )


def tick_specific_data_from_dagster_tick(graphene_info, tick):
    if tick.status == ScheduleTickStatus.SUCCESS:
        run_id = tick.run_id
        run = None
        if graphene_info.context.instance.has_run(run_id):
            run = graphene_info.schema.type_named('PipelineRun')(
                graphene_info.context.instance.get_run_by_id(run_id)
            )
        return graphene_info.schema.type_named('ScheduleTickSuccessData')(run=run)
    elif tick.status == ScheduleTickStatus.FAILURE:
        error = tick.error
        return graphene_info.schema.type_named('ScheduleTickFailureData')(error=error)


class DauphinScheduleTickStatsSnapshot(dauphin.ObjectType):
    class Meta(object):
        name = 'ScheduleTickStatsSnapshot'

    ticks_started = dauphin.NonNull(dauphin.Int)
    ticks_succeeded = dauphin.NonNull(dauphin.Int)
    ticks_skipped = dauphin.NonNull(dauphin.Int)
    ticks_failed = dauphin.NonNull(dauphin.Int)

    def __init__(self, stats):
        super(DauphinScheduleTickStatsSnapshot, self).__init__(
            ticks_started=stats.ticks_started,
            ticks_succeeded=stats.ticks_succeeded,
            ticks_skipped=stats.ticks_skipped,
            ticks_failed=stats.ticks_failed,
        )
        self._stats = check.inst_param(stats, 'stats', ScheduleTickStatsSnapshot)


class DauphinRunningSchedule(dauphin.ObjectType):
    class Meta(object):
        name = 'RunningSchedule'

    schedule_definition = dauphin.NonNull('ScheduleDefinition')
    python_path = dauphin.Field(dauphin.String)
    repository_path = dauphin.Field(dauphin.String)
    status = dauphin.NonNull('ScheduleStatus')
    runs = dauphin.Field(dauphin.non_null_list('PipelineRun'), limit=dauphin.Int())
    runs_count = dauphin.NonNull(dauphin.Int)
    ticks = dauphin.Field(dauphin.non_null_list('ScheduleTick'), limit=dauphin.Int())
    ticks_count = dauphin.NonNull(dauphin.Int)
    stats = dauphin.NonNull('ScheduleTickStatsSnapshot')
    # TODO: Delete attempts and attempts_count in 0.8.0 release
    # https://github.com/dagster-io/dagster/issues/228
    attempts = dauphin.Field(dauphin.non_null_list('ScheduleAttempt'), limit=dauphin.Int())
    attempts_count = dauphin.NonNull(dauphin.Int)
    logs_path = dauphin.NonNull(dauphin.String)

    def __init__(self, graphene_info, schedule):
        self._schedule = check.inst_param(schedule, 'schedule', Schedule)

        super(DauphinRunningSchedule, self).__init__(
            schedule_definition=graphene_info.schema.type_named('ScheduleDefinition')(
                graphene_info=graphene_info,
                schedule_def=get_dagster_schedule_def(graphene_info, schedule.name),
            ),
            status=schedule.status,
            python_path=schedule.python_path,
            repository_path=schedule.repository_path,
        )

    # TODO: Delete in 0.8.0 release
    # https://github.com/dagster-io/dagster/issues/228
    def resolve_attempts(self, graphene_info, **kwargs):
        limit = kwargs.get('limit')

        results = get_schedule_attempt_filenames(graphene_info, self._schedule.name)
        if limit is None:
            limit = len(results)
        latest_results = heapq.nlargest(limit, results, key=os.path.getctime)

        attempts = []
        for result_path in latest_results:
            with open(result_path, 'r') as f:
                line = f.readline()
                if not line:
                    continue  # File is empty

                start_scheduled_execution_response = json.loads(line)
                run = None

                if 'errors' in start_scheduled_execution_response:
                    status = DauphinScheduleAttemptStatus.ERROR
                    json_result = start_scheduled_execution_response['errors']
                else:
                    json_result = start_scheduled_execution_response['data'][
                        'startScheduledExecution'
                    ]
                    typename = json_result['__typename']

                    if (
                        typename == 'StartPipelineExecutionSuccess'
                        or typename == 'LaunchPipelineExecutionSuccess'
                    ):
                        status = DauphinScheduleAttemptStatus.SUCCESS
                        run_id = json_result['run']['runId']
                        if graphene_info.context.instance.has_run(run_id):
                            run = graphene_info.schema.type_named('PipelineRun')(
                                graphene_info.context.instance.get_run_by_id(run_id)
                            )
                    elif typename == 'ScheduledExecutionBlocked':
                        status = DauphinScheduleAttemptStatus.SKIPPED
                    else:
                        status = DauphinScheduleAttemptStatus.ERROR

                attempts.append(
                    graphene_info.schema.type_named('ScheduleAttempt')(
                        time=os.path.getctime(result_path),
                        json_result=json.dumps(json_result),
                        status=status,
                        run=run,
                    )
                )

        return attempts

    # TODO: Delete in 0.8.0 release
    # https://github.com/dagster-io/dagster/issues/228
    def resolve_attempts_count(self, graphene_info):
        attempt_files = get_schedule_attempt_filenames(graphene_info, self._schedule.name)
        return len(attempt_files)

    # TODO: Delete in 0.8.0 release
    # https://github.com/dagster-io/dagster/issues/228
    def resolve_logs_path(self, graphene_info):
        instance = graphene_info.context.instance
        repository = graphene_info.context.get_repository()
        return instance.log_path_for_schedule(repository, self._schedule.name)

    def resolve_stats(self, graphene_info):
        repository = graphene_info.context.get_repository()
        stats = graphene_info.context.instance.get_schedule_tick_stats_by_schedule(
            repository, self._schedule.name
        )
        return graphene_info.schema.type_named('ScheduleTickStatsSnapshot')(stats)

    def resolve_ticks(self, graphene_info, limit=None):

        repository = graphene_info.context.get_repository()

        # TODO: Add cursor limit argument to get_schedule_ticks_by_schedule
        # https://github.com/dagster-io/dagster/issues/2291
        ticks = graphene_info.context.instance.get_schedule_ticks_by_schedule(
            repository, self._schedule.name
        )

        if not limit:
            tick_subset = ticks
        else:
            tick_subset = ticks[:limit]

        return [
            graphene_info.schema.type_named('ScheduleTick')(
                tick_id=tick.tick_id,
                status=tick.status,
                timestamp=tick.timestamp,
                tick_specific_data=tick_specific_data_from_dagster_tick(graphene_info, tick),
            )
            for tick in tick_subset
        ]

    def resolve_ticks_count(self, graphene_info):
        repository = graphene_info.context.get_repository()
        ticks = graphene_info.context.instance.get_schedule_ticks_by_schedule(
            repository, self._schedule.name
        )
        return len(ticks)

    def resolve_runs(self, graphene_info, **kwargs):
        return [
            graphene_info.schema.type_named('PipelineRun')(r)
            for r in graphene_info.context.instance.get_runs(
                filters=PipelineRunsFilter.for_schedule(self._schedule), limit=kwargs.get('limit'),
            )
        ]

    def resolve_runs_count(self, graphene_info):
        return graphene_info.context.instance.get_runs_count(
            filters=PipelineRunsFilter.for_schedule(self._schedule)
        )


class DauphinScheduler(dauphin.ObjectType):
    class Meta(object):
        name = 'Scheduler'

    runningSchedules = dauphin.non_null_list('RunningSchedule')


class RunningScheduleResult(dauphin.ObjectType):
    class Meta(object):
        name = 'RunningScheduleResult'

    schedule = dauphin.NonNull('RunningSchedule')


class DauphinStartScheduleMutation(dauphin.Mutation):
    class Meta(object):
        name = 'StartScheduleMutation'

    class Arguments(object):
        schedule_name = dauphin.NonNull(dauphin.String)

    Output = dauphin.NonNull('RunningScheduleResult')

    def mutate(self, graphene_info, schedule_name):
        repository = graphene_info.context.get_repository()
        instance = graphene_info.context.instance

        schedule = instance.start_schedule(repository, schedule_name)

        return graphene_info.schema.type_named('RunningScheduleResult')(
            schedule=graphene_info.schema.type_named('RunningSchedule')(
                graphene_info, schedule=schedule
            )
        )


class DauphinStopRunningScheduleMutation(dauphin.Mutation):
    class Meta(object):
        name = 'StopRunningScheduleMutation'

    class Arguments(object):
        schedule_name = dauphin.NonNull(dauphin.String)

    Output = dauphin.NonNull('RunningScheduleResult')

    def mutate(self, graphene_info, schedule_name):
        repository = graphene_info.context.get_repository()
        instance = graphene_info.context.instance

        schedule = instance.stop_schedule(repository, schedule_name)

        return graphene_info.schema.type_named('RunningScheduleResult')(
            schedule=graphene_info.schema.type_named('RunningSchedule')(
                graphene_info, schedule=schedule
            )
        )



from graphql.execution.base import ResolveInfo

from dagster import check
from dagster.config.validate import validate_config
from dagster.core.definitions.environment_schema import EnvironmentSchema, create_environment_schema
from dagster.core.definitions.pipeline import ExecutionSelector, PipelineDefinition

from .fetch_pipelines import get_pipeline_def_from_selector
from .utils import UserFacingGraphQLError, capture_dauphin_error


@capture_dauphin_error
def resolve_environment_schema_or_error(graphene_info, selector, mode):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(selector, 'selector', ExecutionSelector)
    check.opt_str_param(mode, 'mode')

    pipeline_def = get_pipeline_def_from_selector(graphene_info, selector)

    if mode is None:
        mode = pipeline_def.get_default_mode_name()

    if not pipeline_def.has_mode_definition(mode):
        raise UserFacingGraphQLError(
            graphene_info.schema.type_named('ModeNotFoundError')(mode=mode, selector=selector)
        )

    return graphene_info.schema.type_named('EnvironmentSchema')(
        dagster_pipeline=pipeline_def,
        environment_schema=create_environment_schema(pipeline_def, mode),
    )


@capture_dauphin_error
def resolve_is_environment_config_valid(
    graphene_info, environment_schema, dagster_pipeline, environment_dict
):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(environment_schema, 'environment_schema', EnvironmentSchema)
    check.inst_param(dagster_pipeline, 'dagster_pipeline', PipelineDefinition)
    check.dict_param(environment_dict, 'environment_dict', key_type=str)

    validated_config = validate_config(environment_schema.environment_type, environment_dict)
    dauphin_pipeline = graphene_info.schema.type_named('Pipeline').from_pipeline_def(
        dagster_pipeline
    )

    if not validated_config.success:
        raise UserFacingGraphQLError(
            graphene_info.schema.type_named('PipelineConfigValidationInvalid')(
                pipeline=dauphin_pipeline,
                errors=[
                    graphene_info.schema.type_named(
                        'PipelineConfigValidationError'
                    ).from_dagster_error(
                        dagster_pipeline.get_config_schema_snapshot(), err,
                    )
                    for err in validated_config.errors
                ],
            )
        )

    return graphene_info.schema.type_named('PipelineConfigValidationValid')(dauphin_pipeline)



import json

from dagster_graphql.cli import execute_query_against_remote
from dagster_graphql.implementation.utils import ExecutionMetadata, ExecutionParams

from dagster import check
from dagster.core.definitions.mode import DEFAULT_MODE_NAME
from dagster.core.definitions.pipeline import ExecutionSelector

from .query import START_PIPELINE_EXECUTION_MUTATION


def execute_remote_pipeline_run(
    host, pipeline_name, environment_dict=None, tags=None, solid_subset=None, mode=None
):
    check.str_param(host, 'host')
    check.str_param(pipeline_name, 'pipeline_name')
    environment_dict = check.opt_dict_param(environment_dict, 'environment_dict', key_type=str)
    check.opt_dict_param(tags, 'tags', key_type=str, value_type=str)
    check.opt_nullable_list_param(solid_subset, 'solid_subset', of_type=str)
    mode = check.opt_str_param(mode, 'mode', DEFAULT_MODE_NAME)

    selector = ExecutionSelector(pipeline_name, solid_subset)
    execution_params = ExecutionParams(
        selector=selector,
        environment_dict=environment_dict,
        mode=mode,
        execution_metadata=ExecutionMetadata(run_id=None, tags=tags or {}),
        step_keys=None,
        previous_run_id=None,
    )

    result = execute_query_against_remote(
        host,
        START_PIPELINE_EXECUTION_MUTATION,
        variables=json.dumps({'executionParams': execution_params.to_graphql_input()}),
    )

    return result



from dagster_graphql.client.query import (
    LAUNCH_PIPELINE_EXECUTION_MUTATION,
    LAUNCH_PIPELINE_REEXECUTION_MUTATION,
)
from dagster_graphql.test.utils import define_context_for_repository_yaml, execute_dagster_graphql

from dagster import seven
from dagster.core.instance import DagsterInstance, InstanceType
from dagster.core.storage.event_log import InMemoryEventLogStorage
from dagster.core.storage.local_compute_log_manager import NoOpComputeLogManager
from dagster.core.storage.root import LocalArtifactStorage
from dagster.core.storage.runs import InMemoryRunStorage
from dagster.core.utils import make_new_run_id
from dagster.utils import file_relative_path

from .execution_queries import (
    START_PIPELINE_EXECUTION_SNAPSHOT_QUERY,
    START_PIPELINE_REEXECUTION_SNAPSHOT_QUERY,
)
from .setup import (
    csv_hello_world_solids_config,
    csv_hello_world_solids_config_fs_storage,
    define_test_context,
)
from .utils import InMemoryRunLauncher

RUN_QUERY = '''
query RunQuery($runId: ID!) {
  pipelineRunOrError(runId: $runId) {
    __typename
    ... on PipelineRun {
        status
      }
    }
  }
'''


def sanitize_result_data(result_data):
    if isinstance(result_data, dict):
        if 'path' in result_data:
            result_data['path'] = 'DUMMY_PATH'
        result_data = {k: sanitize_result_data(v) for k, v in result_data.items()}
    elif isinstance(result_data, list):
        for i in range(len(result_data)):
            result_data[i] = sanitize_result_data(result_data[i])
    else:
        pass
    return result_data


def test_full_pipeline_reexecution_fs_storage(snapshot):
    run_id = make_new_run_id()
    instance = DagsterInstance.ephemeral()
    result_one = execute_dagster_graphql(
        define_test_context(instance=instance),
        START_PIPELINE_EXECUTION_SNAPSHOT_QUERY,
        variables={
            'executionParams': {
                'selector': {'name': 'csv_hello_world'},
                'environmentConfigData': csv_hello_world_solids_config_fs_storage(),
                'executionMetadata': {'runId': run_id},
                'mode': 'default',
            }
        },
    )

    assert (
        result_one.data['startPipelineExecution']['__typename'] == 'StartPipelineExecutionSuccess'
    )

    snapshot.assert_match(sanitize_result_data(result_one.data))

    # reexecution
    new_run_id = make_new_run_id()

    result_two = execute_dagster_graphql(
        define_test_context(instance=instance),
        START_PIPELINE_REEXECUTION_SNAPSHOT_QUERY,
        variables={
            'executionParams': {
                'selector': {'name': 'csv_hello_world'},
                'environmentConfigData': csv_hello_world_solids_config_fs_storage(),
                'executionMetadata': {
                    'runId': new_run_id,
                    'rootRunId': run_id,
                    'parentRunId': run_id,
                },
                'mode': 'default',
            }
        },
    )

    query_result = result_two.data['startPipelineReexecution']
    assert query_result['__typename'] == 'StartPipelineReexecutionSuccess'
    assert query_result['run']['rootRunId'] == run_id
    assert query_result['run']['parentRunId'] == run_id


def test_full_pipeline_reexecution_in_memory_storage(snapshot):
    run_id = make_new_run_id()
    instance = DagsterInstance.ephemeral()
    result_one = execute_dagster_graphql(
        define_test_context(instance=instance),
        START_PIPELINE_EXECUTION_SNAPSHOT_QUERY,
        variables={
            'executionParams': {
                'selector': {'name': 'csv_hello_world'},
                'environmentConfigData': csv_hello_world_solids_config(),
                'executionMetadata': {'runId': run_id},
                'mode': 'default',
            }
        },
    )

    assert (
        result_one.data['startPipelineExecution']['__typename'] == 'StartPipelineExecutionSuccess'
    )

    snapshot.assert_match(sanitize_result_data(result_one.data))

    # reexecution
    new_run_id = make_new_run_id()

    result_two = execute_dagster_graphql(
        define_test_context(instance=instance),
        START_PIPELINE_REEXECUTION_SNAPSHOT_QUERY,
        variables={
            'executionParams': {
                'selector': {'name': 'csv_hello_world'},
                'environmentConfigData': csv_hello_world_solids_config(),
                'executionMetadata': {
                    'runId': new_run_id,
                    'rootRunId': run_id,
                    'parentRunId': run_id,
                },
                'mode': 'default',
            }
        },
    )

    query_result = result_two.data['startPipelineReexecution']
    assert query_result['__typename'] == 'StartPipelineReexecutionSuccess'
    assert query_result['run']['rootRunId'] == run_id
    assert query_result['run']['parentRunId'] == run_id


def test_pipeline_reexecution_successful_launch():
    test_queue = InMemoryRunLauncher()

    with seven.TemporaryDirectory() as temp_dir:
        instance = DagsterInstance(
            instance_type=InstanceType.EPHEMERAL,
            local_artifact_storage=LocalArtifactStorage(temp_dir),
            run_storage=InMemoryRunStorage(),
            event_storage=InMemoryEventLogStorage(),
            compute_log_manager=NoOpComputeLogManager(temp_dir),
            run_launcher=test_queue,
        )

    context = define_context_for_repository_yaml(
        path=file_relative_path(__file__, '../repository.yaml'), instance=instance
    )
    run_id = make_new_run_id()
    result = execute_dagster_graphql(
        context=context,
        query=LAUNCH_PIPELINE_EXECUTION_MUTATION,
        variables={
            'executionParams': {
                'selector': {'name': 'no_config_pipeline'},
                'environmentConfigData': {'storage': {'filesystem': {}}},
                'executionMetadata': {'runId': run_id},
                'mode': 'default',
            }
        },
    )

    assert result.data['launchPipelineExecution']['__typename'] == 'LaunchPipelineExecutionSuccess'
    assert result.data['launchPipelineExecution']['run']['status'] == 'NOT_STARTED'
    test_queue.run_one(instance)
    result = execute_dagster_graphql(context=context, query=RUN_QUERY, variables={'runId': run_id})
    assert result.data['pipelineRunOrError']['__typename'] == 'PipelineRun'
    assert result.data['pipelineRunOrError']['status'] == 'SUCCESS'

    # reexecution
    new_run_id = make_new_run_id()
    result = execute_dagster_graphql(
        context=context,
        query=LAUNCH_PIPELINE_REEXECUTION_MUTATION,
        variables={
            'executionParams': {
                'selector': {'name': 'no_config_pipeline'},
                'environmentConfigData': {'storage': {'filesystem': {}}},
                'executionMetadata': {
                    'runId': new_run_id,
                    'rootRunId': run_id,
                    'parentRunId': run_id,
                },
                'mode': 'default',
            }
        },
    )
    assert (
        result.data['launchPipelineReexecution']['__typename'] == 'LaunchPipelineReexecutionSuccess'
    )

    test_queue.run_one(instance)

    result = execute_dagster_graphql(
        context=context, query=RUN_QUERY, variables={'runId': new_run_id}
    )
    assert result.data['pipelineRunOrError']['__typename'] == 'PipelineRun'
    assert result.data['pipelineRunOrError']['status'] == 'SUCCESS'


def test_pipeline_reexecution_launcher_missing():
    context = define_context_for_repository_yaml(
        path=file_relative_path(__file__, '../repository.yaml')
    )

    run_id = make_new_run_id()
    new_run_id = make_new_run_id()

    result = execute_dagster_graphql(
        context=context,
        query=LAUNCH_PIPELINE_REEXECUTION_MUTATION,
        variables={
            'executionParams': {
                'selector': {'name': 'no_config_pipeline'},
                'executionMetadata': {
                    'runId': new_run_id,
                    'rootRunId': run_id,
                    'parentRunId': run_id,
                },
                'mode': 'default',
            }
        },
    )

    assert result.data['launchPipelineReexecution']['__typename'] == 'RunLauncherNotDefinedError'



from collections import OrderedDict, defaultdict

from dagster_graphql.implementation.fetch_pipelines import get_dauphin_pipeline_from_pipeline_index
from dagster_graphql.schema.solids import (
    DauphinSolidInvocationSite,
    DauphinUsedSolid,
    build_dauphin_solid_handles,
)


def get_solid(graphene_info, name):
    return get_used_solid_map(graphene_info)[name]


def get_solids(graphene_info):
    return get_used_solid_map(graphene_info).values()


def get_used_solid_map(graphene_info):
    repository_index = graphene_info.context.get_repository_index()
    inv_by_def_name = defaultdict(list)
    definitions = []

    for pipeline_index in repository_index.get_pipeline_indices():
        for handle in build_dauphin_solid_handles(
            pipeline_index, pipeline_index.dep_structure_index
        ):
            definition = handle.solid.get_dauphin_solid_definition()
            if definition.name not in inv_by_def_name:
                definitions.append(definition)
            inv_by_def_name[definition.name].append(
                DauphinSolidInvocationSite(
                    pipeline=get_dauphin_pipeline_from_pipeline_index(
                        graphene_info, pipeline_index
                    ),
                    solidHandle=handle,
                )
            )
    return OrderedDict(
        (
            definition.name,
            DauphinUsedSolid(
                definition=definition,
                invocations=sorted(
                    inv_by_def_name[definition.name], key=lambda i: i.solidHandle.handleID
                ),
            ),
        )
        for definition in sorted(definitions, key=lambda d: d.name)
    )



import sys

from dagster_graphql.implementation.context import DagsterGraphQLContext
from dagster_graphql.schema.pipelines import DauphinPipeline, DauphinPipelineSnapshot
from graphql.execution.base import ResolveInfo

from dagster import check
from dagster.core.definitions.pipeline import ExecutionSelector
from dagster.core.errors import DagsterInvalidDefinitionError
from dagster.core.snap.pipeline_snapshot import PipelineIndex
from dagster.utils.error import serializable_error_info_from_exc_info

from .utils import UserFacingGraphQLError, capture_dauphin_error


def get_pipeline_snapshot(graphene_info, snapshot_id):
    check.str_param(snapshot_id, 'snapshot_id')
    pipeline_snapshot = graphene_info.context.instance.get_pipeline_snapshot(snapshot_id)
    return DauphinPipelineSnapshot(PipelineIndex(pipeline_snapshot))


@capture_dauphin_error
def get_pipeline_snapshot_or_error_from_pipeline_name(graphene_info, pipeline_name):
    check.str_param(pipeline_name, 'pipeline_name')
    pipeline_def = get_pipeline_definition(graphene_info, pipeline_name)
    return DauphinPipelineSnapshot(pipeline_def.get_pipeline_index())


@capture_dauphin_error
def get_pipeline_snapshot_or_error_from_snapshot_id(graphene_info, snapshot_id):
    check.str_param(snapshot_id, 'snapshot_id')
    return _get_dauphin_pipeline_snapshot_from_instance(graphene_info.context.instance, snapshot_id)


# extracted this out to test
def _get_dauphin_pipeline_snapshot_from_instance(instance, snapshot_id):
    from dagster_graphql.schema.errors import DauphinPipelineSnapshotNotFoundError

    if not instance.has_pipeline_snapshot(snapshot_id):
        raise UserFacingGraphQLError(DauphinPipelineSnapshotNotFoundError(snapshot_id))

    pipeline_snapshot = instance.get_pipeline_snapshot(snapshot_id)

    if not pipeline_snapshot:
        # Either a temporary error or it has been deleted in the interim
        raise UserFacingGraphQLError(DauphinPipelineSnapshotNotFoundError(snapshot_id))

    return DauphinPipelineSnapshot(PipelineIndex(pipeline_snapshot))


@capture_dauphin_error
def get_pipeline_or_error(graphene_info, selector):
    '''Returns a DauphinPipelineOrError.'''
    return get_dauphin_pipeline_from_selector(graphene_info, selector)


def get_pipeline_or_raise(graphene_info, selector):
    '''Returns a DauphinPipeline or raises a UserFacingGraphQLError if one cannot be retrieved
    from the selector, e.g., the pipeline is not present in the loaded repository.'''
    return get_dauphin_pipeline_from_selector(graphene_info, selector)


def get_pipeline_reference_or_raise(graphene_info, selector):
    '''Returns a DauphinPipelineReference or raises a UserFacingGraphQLError if a pipeline
    reference cannot be retrieved from the selector, e.g, a UserFacingGraphQLError that wraps an
    InvalidSubsetError.'''
    return get_dauphin_pipeline_reference_from_selector(graphene_info, selector)


def get_dauphin_pipeline_reference_from_selector(graphene_info, selector):
    from ..schema.errors import DauphinPipelineNotFoundError, DauphinInvalidSubsetError

    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(selector, 'selector', ExecutionSelector)
    try:
        return get_dauphin_pipeline_from_selector(graphene_info, selector)
    except UserFacingGraphQLError as exc:
        if (
            isinstance(exc.dauphin_error, DauphinPipelineNotFoundError)
            or
            # At this time DauphinPipeline represents a potentially subsetted
            # pipeline so if the solids used to subset no longer exist
            # we can't return the correct instance so we fallback to
            # UnknownPipeline
            isinstance(exc.dauphin_error, DauphinInvalidSubsetError)
        ):
            return graphene_info.schema.type_named('UnknownPipeline')(selector.name)

        raise


@capture_dauphin_error
def get_pipelines_or_error(graphene_info):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    return _get_pipelines(graphene_info)


def get_pipelines_or_raise(graphene_info):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    return _get_pipelines(graphene_info)


def _get_pipelines(graphene_info):
    dauphin_pipelines = get_dauphin_pipelines(graphene_info)
    return graphene_info.schema.type_named('PipelineConnection')(
        nodes=sorted(dauphin_pipelines, key=lambda pipeline: pipeline.name)
    )


def get_dauphin_pipelines(graphene_info):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    repository_index = graphene_info.context.get_repository_index()
    return [
        get_dauphin_pipeline_from_pipeline_index(graphene_info, pipeline_index)
        for pipeline_index in repository_index.get_pipeline_indices()
    ]


def get_dauphin_pipeline_from_pipeline_index(graphene_info, pipeline_index):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(pipeline_index, 'pipeline_index', PipelineIndex)
    if isinstance(graphene_info.context, DagsterGraphQLContext):
        pipeline = get_pipeline_definition(graphene_info, pipeline_index.pipeline_snapshot.name)
        return DauphinPipeline(pipeline_index, presets=pipeline.get_presets())
    return DauphinPipeline(pipeline_index)


def get_dauphin_pipeline_from_selector(graphene_info, selector):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(selector, 'selector', ExecutionSelector)

    if isinstance(graphene_info.context, DagsterGraphQLContext):
        pipeline_definition = get_pipeline_def_from_selector(graphene_info, selector)
        return DauphinPipeline.from_pipeline_def(pipeline_definition)

    # TODO: Support solid sub selection.
    check.invariant(
        not selector.solid_subset,
        desc="DagsterSnapshotGraphQLContext doesn't support pipeline sub-selection.",
    )

    repository_index = graphene_info.context.get_repository_index()
    if not repository_index.has_pipeline_index(selector.name):
        raise UserFacingGraphQLError(
            graphene_info.schema.type_named('PipelineNotFoundError')(pipeline_name=selector.name)
        )
    return DauphinPipeline(repository_index.get_pipeline_index(selector.name))


def get_pipeline_def_from_selector(graphene_info, selector):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(selector, 'selector', ExecutionSelector)
    if not selector.solid_subset:
        return get_pipeline_definition(graphene_info, selector.name)
    return get_solid_subset_pipeline_definition(graphene_info, selector)


def get_pipeline_definition(graphene_info, pipeline_name):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    repository = graphene_info.context.get_repository()
    if not repository.has_pipeline(pipeline_name):
        raise UserFacingGraphQLError(
            graphene_info.schema.type_named('PipelineNotFoundError')(pipeline_name=pipeline_name)
        )
    return graphene_info.context.get_pipeline(pipeline_name)


def get_solid_subset_pipeline_definition(graphene_info, selector):
    orig_pipeline = get_pipeline_definition(graphene_info, selector.name)
    for solid_name in selector.solid_subset:
        if not orig_pipeline.has_solid_named(solid_name):
            raise UserFacingGraphQLError(
                graphene_info.schema.type_named('InvalidSubsetError')(
                    message='Solid "{solid_name}" does not exist in "{pipeline_name}"'.format(
                        solid_name=solid_name, pipeline_name=selector.name
                    ),
                    pipeline=graphene_info.schema.type_named('Pipeline').from_pipeline_def(
                        orig_pipeline
                    ),
                )
            )
    try:
        return orig_pipeline.build_sub_pipeline(selector.solid_subset)
    except DagsterInvalidDefinitionError:
        raise UserFacingGraphQLError(
            graphene_info.schema.type_named('InvalidSubsetError')(
                message=serializable_error_info_from_exc_info(sys.exc_info()).message,
                pipeline=graphene_info.schema.type_named('Pipeline').from_pipeline_def(
                    orig_pipeline
                ),
            )
        )



from __future__ import absolute_import

from graphql.execution.base import ResolveInfo

from dagster import check
from dagster.core.execution.api import create_execution_plan

from ..fetch_pipelines import get_pipeline_def_from_selector
from ..fetch_runs import get_validated_config
from ..utils import ExecutionMetadata, ExecutionParams, capture_dauphin_error
from .utils import _check_start_pipeline_execution_errors, _create_pipeline_run


@capture_dauphin_error
def launch_pipeline_reexecution(graphene_info, execution_params):
    return _launch_pipeline_execution(graphene_info, execution_params, is_reexecuted=True)


@capture_dauphin_error
def launch_pipeline_execution(graphene_info, execution_params):
    return _launch_pipeline_execution(graphene_info, execution_params)


def _launch_pipeline_execution(graphene_info, execution_params, is_reexecuted=False):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(execution_params, 'execution_params', ExecutionParams)

    if is_reexecuted:
        # required fields for re-execution
        execution_metadata = check.inst_param(
            execution_params.execution_metadata, 'execution_metadata', ExecutionMetadata
        )
        check.str_param(execution_metadata.root_run_id, 'root_run_id')
        check.str_param(execution_metadata.parent_run_id, 'parent_run_id')

    error_type = 'RunLauncherNotDefinedError'
    success_type = (
        'LaunchPipelineExecutionSuccess'
        if not is_reexecuted
        else 'LaunchPipelineReexecutionSuccess'
    )
    instance = graphene_info.context.instance
    run_launcher = instance.run_launcher

    if run_launcher is None:
        return graphene_info.schema.type_named(error_type)()

    pipeline_def = get_pipeline_def_from_selector(graphene_info, execution_params.selector)

    get_validated_config(
        pipeline_def,
        environment_dict=execution_params.environment_dict,
        mode=execution_params.mode,
    )

    execution_plan = create_execution_plan(
        pipeline_def, execution_params.environment_dict, mode=execution_params.mode,
    )

    _check_start_pipeline_execution_errors(graphene_info, execution_params, execution_plan)

    run = instance.launch_run(_create_pipeline_run(instance, pipeline_def, execution_params))

    return graphene_info.schema.type_named(success_type)(
        run=graphene_info.schema.type_named('PipelineRun')(run)
    )
